package saboteur;

/**
 *
 * Authors: Dyon Veldhuis, Eric Jansen, Eveline Broers
 */

import java.util.*;
import java.lang.Math.*;

public class Agent {
    int id;
    boolean saboteur;
    public Vector<Card> handCards = new Vector<Card>();  
    // shadowCards to save a 180 degrees rotated version of the normal cards
    public Vector<Card> shadowHandCards = new Vector<Card>(); 
    KripkeModel model;
    int treshold;
    private boolean blocked;
    private int[] knowGoal;
    private int gold;
    private Deceive deceive;    
    Card discardCard = new Card();
    private boolean explicit = false;
    
    public Agent(boolean saboteur, int id, int round) {               
        this.id = id;
        this.saboteur = saboteur;
        treshold = 4;
        blocked = false;
        knowGoal = new int[]{0,0,0,0};
        //System.out.println("round: " + round + " agent: " + id + " gold: " + gold);
        deceive = new Deceive(explicit);
        model = new KripkeModel(id, saboteur);
    }
    
    public Turn takeTurn(Agent[] agents, PathCard[][] board) {
        for(int i = 0; i < handCards.size(); i++) {
            if(handCards.get(i) instanceof PathCard) {
                PathCard shadowCard = (PathCard)handCards.get(i);
                boolean[] sides = new boolean[4];
                boolean[] opened = new boolean[4];
                for(int j = 0; j < 4; j++){
                    sides[j] = shadowCard.getSides()[(j+2)%4];
                    opened[j] = shadowCard.getOpened()[(j+2)%4];
                }
                shadowHandCards.add(new PathCard(sides, opened));
            } else {
                shadowHandCards.add(handCards.get(i));
            }
        }
        
        double bestValue[] = new double[]{0,0,0};
        bestValue[0] = treshold;
        int bestCard = 0;
        Turn turn;
        boolean turned = false;
        
        if(saboteur){
            double[] bestSaboteur = deceive.takeTurn(agents, blocked, treshold, board, handCards, shadowHandCards, model, id, knowGoal);
            System.arraycopy(bestSaboteur, 0, bestValue, 0, 3);
            bestCard = (int)bestSaboteur[3];
            if(bestSaboteur[4] > 1){
                turned = true;
            }
        } else { 

            
            for(int i = 0; i < handCards.size(); i++) {
                double currentValue[] = evaluate(handCards.get(i), agents, board);
                if(currentValue[0] > bestValue[0]){
                    bestValue = currentValue;
                    bestCard = i;
                }
            }
            for(int i = 0; i < shadowHandCards.size(); i++) {
                double currentValue[] = evaluate(shadowHandCards.get(i), agents, board);
                if(currentValue[0] > bestValue[0]){
                    bestValue = currentValue;
                    bestCard = i;
                    turned = true;
                }
            }
        }
        Card playCard;
        if(bestValue[0] > treshold){
            if(turned){
                playCard = shadowHandCards.get(bestCard);
            } else{
                playCard = handCards.get(bestCard);
            }
            handCards.remove(bestCard);
            shadowHandCards.clear();
        } else if(saboteur){
            int removeCard = deceive.discard(agents, blocked, treshold, board, handCards, shadowHandCards, model, id, knowGoal);
            discardCard = handCards.get(removeCard);
            handCards.remove(removeCard);
            shadowHandCards.clear();
            return null;
        } else {
            discard(model, agents, board);
            shadowHandCards.clear();
            return null;
        }
        int[] target = new int[2];
        target[0] = (int)bestValue[1];
        target[1] = (int)bestValue[2];
        turn = new Turn(playCard, target, (int)bestValue[0]);
        return turn;
    }
   
    private void discard(KripkeModel model, Agent[] agents, PathCard[][] board){
        double worstValue = 10;
        int worstCard = 0;
        for(int i = 0; i < handCards.size(); i++) {
            double[] currentValue = evaluate(handCards.get(i), agents, board);
            if(currentValue[0] < worstValue){
                worstValue = currentValue[0];
                worstCard = i;
            }
        }
        //System.out.println(id + " just discarded a " + handCards.get(worstCard).getClass() + " from his hand with value " + worstValue);
        discardCard = handCards.get(worstCard);
        handCards.remove(worstCard);
    }
    
    //evaluate the strength of a card
    private double[] evaluate(Card card, Agent[] agents, PathCard[][] board){
        return evaluate(card,agents,false,board);
    }
    
    private double[] evaluate(Card card, Agent[] agents,boolean discard, PathCard[][] board) {
        //first number is the numeric evaluation of a card, the second number is either the id of an agent
        //or combined with the third number the coordinates of a pathcard
        double evaluation[] = new double[]{0,0,0};
        
        if(card instanceof PathCard){
            if(blocked == false || discard){
                evaluation = calculatePath(card, board);
            } 
            
        } else if (card instanceof BlockCard){
            double suspicion[] = model.getSuspicion(id);//find out which agent is most susicious
            if(agents[(int)suspicion[1]].getBlocked() == false){//make sure that agent isn already blocked
                evaluation[0] = ((suspicion[0]/suspicion[2])*10)-2;//the likelihood of this suspicion
                evaluation[1] = suspicion[1];
            } else {
                suspicion = model.getSuspicion(id,(int)suspicion[0]);
                if(agents[(int)suspicion[1]].getBlocked() == false){
                    evaluation[0] = ((suspicion[0]/suspicion[2])*10)-2;
                    evaluation[1] = suspicion[1];
                }
            }
            
        } else if (card instanceof DeblockCard){//evaluation for deblockcard
            if(blocked){//if you are blocked yourself, you want to deblock yourself
                evaluation[0] = 8;
                evaluation[1] = id;
            } else {
                double suspicion[] = model.getLeastSuspicious(id);//find out which agent is most susicious
                if(agents[(int)suspicion[1]].getBlocked()){//make sure that agent isn already blocked
                    evaluation[0] = ((suspicion[0]/suspicion[2])*10)-2;//the likelihood of this suspicion
                    evaluation[1] = suspicion[1];
                } else {
                    suspicion = model.getLeastSuspicious(id,(int)suspicion[0]);
                    if(agents[(int)suspicion[1]].getBlocked()){
                        evaluation[0] = ((suspicion[0]/suspicion[2])*10)-2;
                        evaluation[1] = suspicion[1];
                    }
                }
                int maybeSaboteur = (int)model.getSuspicion(id)[1];
                int maybeSaboteur2 = (int)model.getSuspicion(id, maybeSaboteur)[1];
                for(int i = 1; i < 5; i++) {
                    int check = (i + id) % 5;
                    if(check != maybeSaboteur && check != maybeSaboteur2 && agents[check].getBlocked()) {
                        evaluation[0] = 7;
                        evaluation[1] = check;
                    }
                }
            }
            
            
        } else if (card instanceof DemolishCard){
            double currentValue = 0;
            for(int i = 0; i < 13; i++){
                for(int j = 0; j < 13; j++){
                    if(board[i][j] != null && !(board[i][j] instanceof GoalCard) && !(board[i][j] instanceof StartCard)){
                        currentValue = calculateDeleteScore(i,j,board[i][j],board);
                        if(currentValue > evaluation[0]){
                            evaluation[0] = currentValue;
                            evaluation[1] = i;
                            evaluation[2] = j;
                        }
                    }
                }
            }
        } else { //evaluation of viewcard
            if(knowGoal[0]==0){ //check if you know where the gold is
                evaluation[0] = 9; //if you don't know where to go to, you really want to know!
                evaluation[2] = 10; //goalcards always have y-coordinate 10
                Random rand = new Random();
                int checkGoal = rand.nextInt(3) + 1;
                if(knowGoal[checkGoal]==0){//check if you know the first goalcard
                    evaluation[1]=(checkGoal * 2) + 2;//if not, your evaluation will return x-coordinate of first goalcard
                } else {
                    evaluation[1]=(((checkGoal % 3) + 1) * 2) + 2;//if you know the first, check the second goalcard
                }
            }
        }
            
        return evaluation;
    }
    
    public int getID(){
        return id;
    }
    
    public boolean getBlocked(){
        return blocked;
    }
    
    public void setBlocked(boolean blocker){
        blocked = blocker;
    }
    
    public void setRole(boolean role){
        saboteur = role;
    }
    
    public boolean getRole(){
        return saboteur;
    }
    
    public void resetKripke(){
        model = new KripkeModel(id, saboteur);
    }
    
    public int setGoal(int goalNumber, boolean gold){
        if(gold){//if gold was struck, set that goalcard as goal
            knowGoal[0] = 1;
            knowGoal[goalNumber+1] = 1;
            return goalNumber;
        } else {//else take note that the seen goalcard holds a rock
            knowGoal[goalNumber+1] = 2;
            if(knowGoal[1]==2){// if you've seen two rocks, you know the third one is your goal
                if(knowGoal[2]==2){
                    knowGoal[3]=1;
                    knowGoal[0]=1;
                    return 2;
                } else if(knowGoal[3]==2){
                    knowGoal[2]=1;
                    knowGoal[0]=1;
                    return 1;
                }
            } else if(knowGoal[2]==2 && knowGoal[3]==2){
                knowGoal[1]=1;
                knowGoal[0]=1;
                return 0;
            }
        } 
        return 10;
    }
    
    public void resetGoal(){
        for(int i = 0; i < 4; i++){
            knowGoal[i] = 0;
        }
    }
    
    public int knowsGoal(int i){
        if(saboteur){
            return deceive.knowsGoal(knowGoal[i]);
        }
        return knowGoal[i]; 
    }
    
    public int getGoal(){
        if(saboteur){
            deceive.getGoal(knowGoal);
        }
        return getRealGoal();
    }
    public int getRealGoal(){
        if(knowGoal[0] == 0){
            return 0;
        }
        if(knowGoal[1] == 1){
            return 1;
        }
        if(knowGoal[2] == 1){
            return 2;
        }
        return 3;            
    }
    
    
    private double[] calculatePath(Card card, PathCard[][] board){
        PathCard pathcard = (PathCard)card;
        boolean above = false;
        boolean below = false;
        boolean left = false;
        boolean right = false;
        boolean sidesfit = false;
        double currentValue;
        double[] returnValue = new double[]{-1.0,-1.0,-1.0};
        
        
        for(int i = 0; i < 13; i++){
            for(int j = 0; j < 13; j++){
                if(board[i][j] == null){
                    if(i < 12){
                        left = checkConnection(pathcard, board[i+1][j], 3);
                    } else {
                        left = true;
                    }
                    if(i > 0){
                        right = checkConnection(pathcard, board[i-1][j], 1);
                    } else {
                        right = true;
                    }
                    if(j < 12){
                        above = checkConnection(pathcard, board[i][j+1], 0);
                    } else {
                        above = true;
                    }
                    if(j > 0){
                        below = checkConnection(pathcard, board[i][j-1], 2);
                    } else {
                        below = true;
                    }
                    
                    if(above && below && left && right){
                        sidesfit = true;
                    }
                    boolean[][] shadowBoard = new boolean[13][13];
                    for(int o = 0; o < 13; o++) {
                        for(int p = 0; p < 13; p++) {
                            shadowBoard[o][p] = false;
                        }
                    }
                    board[i][j] = pathcard;
                    if(sidesfit && linksToStart(board, i, j, shadowBoard)){
                        currentValue = calculatePathScore(i,j,pathcard, board);
                        if(currentValue > returnValue[0]){
                            returnValue[0] = currentValue;
                            returnValue[1] = i;
                            returnValue[2] = j;
                            //System.out.println("x: " + i + ", y: " + j + ", score: " + currentValue);
                            //System.err.println(pathcard.getSides()[0] + " " + pathcard.getSides()[1] + " " + pathcard.getSides()[2] + " " + pathcard.getSides()[3]);
                        }
                    }
                    board[i][j] = null;
                }
                sidesfit = false;
                above = false;
                below = false;
                left = false;
                right = false;
            }
        }
        int[] returnInt = new int[3];
        for(int i = 0; i < 3; i++){
            returnInt[i] = (int)returnValue[i];
        }
        return returnValue;
    }
    
    private boolean checkConnection(PathCard card, PathCard boardcard, int goalside){
        if(boardcard == null){
            return true;
        }
        if(card.getSides()[goalside] == boardcard.getSides()[(goalside+2)%4]){
            return true;
        }  
        return false;
    }

    private double calculatePathScore(int x, int y, PathCard card, PathCard[][] board) {
        double cardScore = 0;
        int goalX = 6;
        double distance;
        int branches = 0;
        boolean goodX = false;
        boolean goodY = false;
        if(knowGoal[0] == 1){
            for(int i = 1; i < 4; i++){
                if(knowGoal[i] == 1){
                    goalX = (2*i)+2;
                }
            }
        } 
        for(int i = 0; i < 4; i++){
            if(card.getOpened()[i]){
                branches++;
            }
        }
        if(card.getOpened()[1] && x > 0 && board[x-1][y] instanceof GoalCard){ 
            if(knowGoal[0] == 1 && (x-1) == goalX){
                return 10;
            } else if(knowGoal[((x-1)/2)-2] != 2){
                if(branches > 2){
                    return 10;
                } else{
                    return 9;
                }
            }
        }
        if(card.getOpened()[3] && x < 12 && board[x+1][y] instanceof GoalCard){ 
            if(knowGoal[0] == 1 && (x+1) == goalX){
                return 10;
            } else if(knowGoal[((x+1)/2)-2] != 2){
                if(branches > 2){
                    return 10;
                } else{
                    return 9;
                }
            }
        }
        if(card.getOpened()[2] && y > 0 && board[x][y-1] instanceof GoalCard){
            goodY = true;
        } 
        if(card.getOpened()[0] && y < 12 && board[x][y+1] instanceof GoalCard){
            return 10;
        }
        distance = Math.abs(goalX-x) + Math.abs(y-10);
        
        if(x > goalX && card.getOpened()[1] && board[x-1][y] == null){ 
            goodX = true;
        }
        if(x < goalX && card.getOpened()[3] && board[x+1][y] == null){
            goodX = true;
        }
        if(y > 10 && card.getOpened()[2] && board[x][y-1] == null){
            goodY = true;
        } 
        if(y < 10 && card.getOpened()[0] && board[x][y+1] == null){
            goodY = true;
        }
        if(Math.abs(goalX - x) > Math.abs(10 - y)){
            if(goodX){
                cardScore += 5;
            } else if(goodY){
                cardScore += 4;
            }
            if(goodX && goodY){
                cardScore += 0.5;
            }
        } else {
            if(goodY){
                cardScore += 5;
            } else if(goodX){
                cardScore += 4;
            }
            if(goodX && goodY){
                cardScore += 0.5;
            }
        }
        
        cardScore += branches-1;
        cardScore += (19-distance)/9;
        double print = (19.0-distance)/9.0;
        //System.out.println("added cardscore for distance: " + print);
        
        return cardScore;
    }
    
    private double calculateDeleteScore(int x,int y,PathCard card,PathCard[][] board){
        double cardScore = 0;
        int goalX = 6;
        double distance;
        int branches = 0;
        if(knowGoal[0] == 1){
            for(int i = 1; i < 4; i++){
                if(knowGoal[i] == 1){
                    goalX = (2*i)+2;
                }
            }
        } 
        distance = Math.abs(goalX-x) + Math.abs(y-10);
        for(int i = 0; i < 4; i++){
            if(card.getOpened()[i]){
                branches++;
            }
        }
        cardScore += 8 - 2*branches;
        cardScore += ((19 - distance)/9);
        return cardScore;
    }

    private boolean linksToStart(PathCard[][] board, int xCoordinate, int yCoordinate, boolean[][] shadowBoard) {
        boolean[] opened = board[xCoordinate][yCoordinate].getOpened();
        boolean[] sides = board[xCoordinate][yCoordinate].getSides();
        boolean[] links = new boolean[4];
        shadowBoard[xCoordinate][yCoordinate] = true;//remember that recursion already evaluated this tile
        if(xCoordinate == 6 && yCoordinate == 2) {
            return true;
        } else {           
            if(yCoordinate < 12){
                if(opened[0] && sides[0] && board[xCoordinate][yCoordinate+1]!=null && shadowBoard[xCoordinate][yCoordinate+1]==false) {//check if there is a card there              
                    links[0] = linksToStart(board, xCoordinate, yCoordinate+1, shadowBoard);//Card at North
                }
            }
            if(yCoordinate > 0){
                if(opened[2] && sides[2] && board[xCoordinate][yCoordinate-1]!=null && shadowBoard[xCoordinate][yCoordinate-1]==false) {//check if there is a card there           
                    links[1] = linksToStart(board, xCoordinate, yCoordinate-1, shadowBoard);//Card at South
                }  
            }
            if(xCoordinate > 0){
                if(opened[1] && sides[1] && board[xCoordinate-1][yCoordinate]!=null && shadowBoard[xCoordinate-1][yCoordinate]==false) {//check if there is a card there
                    links[2] = linksToStart(board, xCoordinate-1, yCoordinate, shadowBoard);//Card at Right
                }
            }
            if(xCoordinate < 12){
                if(opened[3] && sides[3] && board[xCoordinate+1][yCoordinate]!=null && shadowBoard[xCoordinate+1][yCoordinate]==false) {//check if there is a card there
                    links[3] = linksToStart(board, xCoordinate+1, yCoordinate, shadowBoard);//Card at Left
                }  
            }
        } 
        if(links[0]||links[1]||links[2]||links[3]){
            return true;
        }
        return false;
    }
    
    public void addGold(){
        gold++;
    }
    
    public int getGold(){
        return gold;
    }
    
    public KripkeModel getKripke(){
        return model;
    }
    
    public void updateKripke(Turn turn, int agentPlayed, PathCard[][] board, Agent[] agents){
        model.update(turn, agentPlayed, board, agents);
        if(saboteur && !deceive.getExplicit()){
            deceive.checkExplicit(board, model, id);
        }
        if(!saboteur){
            model.updateSecondOrderBeliefs(id);
        }
    }
}