 /*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package saboteur;

import java.util.*;

/**
 *
 * @author a1712543
 */
public class Deceive {
    private boolean explicit; //True if you're not afraid to sabotage openly
    private int tresholdHigh;
    public Deceive(boolean explicit){
        this.explicit = explicit;
    }
    
    
    public double[] takeTurn(Agent[] agents, boolean blocked, int treshold, PathCard[][] board, Vector<Card> handCards, Vector<Card> shadowHandCards, KripkeModel model, int id, int[] knowGoal){
        double bestValue[] = new double[]{0,0,0};
        bestValue[0] = treshold;
        int bestCard = 0;
        double turned = 0.0;
        tresholdHigh = treshold + 2;
        
        
        for(int i = 0; i < handCards.size(); i++) {
            double currentValue[] = evaluate(handCards.get(i), agents, board, blocked, model, id, knowGoal);
            if(currentValue[0] > bestValue[0]){
                bestValue = currentValue;
                bestCard = i;
            }
        }
        for(int i = 0; i < shadowHandCards.size(); i++) {
            double currentValue[] = evaluate(shadowHandCards.get(i), agents, board, blocked, model, id, knowGoal);
            if(currentValue[0] > bestValue[0]){
                bestValue = currentValue;
                bestCard = i;
                turned = 10.0;
            }
        }
        
        double[] returnValue = new double[5];
        System.arraycopy(bestValue, 0, returnValue, 0, 3);
        returnValue[3] = bestCard;
        returnValue[4] = turned;
        
        return returnValue; 
    }
    
       
    private double[] evaluate(Card card, Agent[] agents, PathCard[][] board, boolean blocked, KripkeModel model, int id, int[] knowGoal) {
        //first number is the numeric evaluation of a card, the second number is either the id of an agent
        //or combined with the third number the coordinates of a pathcard
        double evaluation[] = new double[]{0,0,0};
        tresholdHigh = 8; //Treshold so you won't play great cards while faking to be a miner
        
        if(card instanceof PathCard){
            if(blocked == false){
                evaluation = calculatePath(card, board, knowGoal);
            } 
            
        } else if (card instanceof BlockCard){
            if(explicit){
                double suspicion[] = model.getSuspicion(id);//find out which agent is most suspicious
                for(int i = 1; i < 5; i++){
                    if((id + i) % 5 != suspicion[1]  && agents[(id + i) % 5].getBlocked() == false){//make sure that agent is not already blocked
                        evaluation[0] = 8;//the likelihood of this suspicion
                        evaluation[1] = (id + i) % 5;
                    }
                }                     
            } else {
                evaluation[0] = 0;
                evaluation[1] = 0;
            }
                
            
        } else if (card instanceof DeblockCard){//evaluation for deblockcard
            if(blocked){//if you are blocked yourself, you want to deblock yourself
                if(explicit){//|| you think others think you are suspicious) {
                    evaluation[0] = 5; // it is very likely that you will immediately be blocked again anyway
                    evaluation[1] = id;
                } else {
                    evaluation[0] = 8;
                    evaluation[1] = id; 
                } 
            }
            
            
        } else if (card instanceof DemolishCard){
            double currentValue = 0;
            if(explicit){    
                for(int i = 0; i < 13; i++){
                    for(int j = 0; j < 13; j++){
                        if(board[i][j] != null && !(board[i][j] instanceof GoalCard) && !(board[i][j] instanceof StartCard)){
                            currentValue = calculateDeleteScore(i,j,board[i][j],board,knowGoal);
                            if(currentValue > evaluation[0]){
                                evaluation[0] = currentValue;
                                evaluation[1] = i;
                                evaluation[2] = j;
                            }
                        }
                    }
                }
            }
        } else { //evaluation of viewcard
            if(knowGoal[0]==0){ //check if you know where the gold is
                evaluation[0] = 9; //if you don't know where to go to, you really want to know!
                evaluation[2] = 10; //goalcards always have y-coordinate 10
                Random rand = new Random();
                int checkGoal = rand.nextInt(3) + 1;
                if(knowGoal[checkGoal]==0){//check if you know the first goalcard
                    evaluation[1]=(checkGoal * 2) + 2;//if not, your evaluation will return x-coordinate of first goalcard
                } else {
                    evaluation[1]=(((checkGoal % 3) + 1) * 2) + 2;//if you know the first, check the second goalcard
                }
            }
        }
        
        return evaluation;
    }
    
    public int discard(Agent[] agents, boolean blocked, int treshold, PathCard[][] board, Vector<Card> handCards, Vector<Card> shadowHandCards, KripkeModel model, int id, int[] knowGoal){
        double worstValue = 10;
        int worstCard = 0;
        double[] currentValue;
        for(int i = 0; i < handCards.size(); i++) {
            currentValue = discardEvaluate(handCards.get(i), agents, board, blocked, model, id, knowGoal);
            if(currentValue[0] < worstValue){
                worstValue = currentValue[0];
                worstCard = i;
            }
        }
        return worstCard;
    }
    
    private double[] discardEvaluate(Card card, Agent[] agents, PathCard[][] board, boolean blocked, KripkeModel model, int id, int[] knowGoal) {
        //first number is the numeric evaluation of a card, the second number is either the id of an agent
        //or combined with the third number the coordinates of a pathcard
        double evaluation[] = new double[]{0,0,0};
        
        
        if(card instanceof PathCard){
            PathCard pathCard = (PathCard) card;
            int branches = 0;
            for(int i = 0; i < 4; i++){
                if(pathCard.getOpened()[i]){
                    branches++;                    
                }
            }
            evaluation[0] = (9-2*branches);
        } else if (card instanceof BlockCard){
            evaluation[0] = 7;//TODO: relatively high value, because you probably don't want to discard this card
            evaluation[1] = 0;                                          
        } else if (card instanceof DeblockCard){//evaluation for deblockcard
            if(explicit){
                evaluation[0] = 6;
            } else {
                evaluation[0] = 8;
            }
        } else if (card instanceof DemolishCard){
            evaluation[0] = 8;
        } else { //evaluation of viewcard
            if(knowGoal[0]==0){ //check if you know where the gold is
                evaluation[0] = 9; //if you don't know where to go to, you really want to know!
            } else {
                evaluation[0] = 1;
            }
        }
        
        return evaluation;
    }
    
    private double[] calculatePath(Card card, PathCard[][] board, int[] knowGoal){
        PathCard pathcard = (PathCard)card;
        boolean above = false;
        boolean below = false;
        boolean left = false;
        boolean right = false;
        boolean sidesfit = false;
        double currentValue;
        double[] returnValue = new double[]{-1.0,-1.0,-1.0};
        
        
        for(int i = 0; i < 13; i++){
            for(int j = 0; j < 13; j++){
                if(board[i][j] == null){
                    if(i < 12){
                        left = checkConnection(pathcard, board[i+1][j], 3);
                    } else {
                        left = true;
                    }
                    if(i > 0){
                        right = checkConnection(pathcard, board[i-1][j], 1);
                    } else {
                        right = true;
                    }
                    if(j < 12){
                        above = checkConnection(pathcard, board[i][j+1], 0);
                    } else {
                        above = true;
                    }
                    if(j > 0){
                        below = checkConnection(pathcard, board[i][j-1], 2);
                    } else {
                        below = true;
                    }
                    
                    if(above && below && left && right){
                        sidesfit = true;
                    }
                    boolean[][] shadowBoard = new boolean[13][13];
                    for(int o = 0; o < 13; o++) {
                        for(int p = 0; p < 13; p++) {
                            shadowBoard[o][p] = false;
                        }
                    }
                    board[i][j] = pathcard;
                    if(sidesfit && linksToStart(board, i, j, shadowBoard)){
                        currentValue = calculatePathScore(i,j,pathcard, board, knowGoal);
                        if(currentValue > returnValue[0]){
                            returnValue[0] = currentValue;
                            returnValue[1] = i;
                            returnValue[2] = j;
                            //System.out.println("x: " + i + ", y: " + j + ", score: " + currentValue);
                            //System.err.println(pathcard.getSides()[0] + " " + pathcard.getSides()[1] + " " + pathcard.getSides()[2] + " " + pathcard.getSides()[3]);
                        }
                    }
                    board[i][j] = null;
                }
                sidesfit = false;
                above = false;
                below = false;
                left = false;
                right = false;
            }
        }
        int[] returnInt = new int[3];
        for(int i = 0; i < 3; i++){
            if(explicit || i > 0 || (int)returnValue[0] < tresholdHigh){
                returnInt[i] = (int)returnValue[i];
            } else {
                returnInt[0] = 0;
            }
        }
        return returnValue;
    }
    
    private boolean checkConnection(PathCard card, PathCard boardcard, int goalside){
        if(boardcard == null){
            return true;
        }
        if(card.getSides()[goalside] == boardcard.getSides()[(goalside+2)%4]){
            return true;
        }  
        return false;
    }

    private double calculatePathScore(int x, int y, PathCard card, PathCard[][] board, int[] knowGoal) {
        double cardScore = 0;
        int goalX = 6;
        double distance;
        int branches = 0;
        boolean goodX = false;
        boolean goodY = false;
        if(knowGoal[0] == 1){
            for(int i = 1; i < 4; i++){
                if(knowGoal[i] == 1){
                    goalX = (2*i)+2;
                }
            }
        } 
        for(int i = 0; i < 4; i++){
            if(card.getOpened()[i]){
                branches++;
            }
        }
        distance = Math.abs(goalX-x) + Math.abs(y-10);
        if(!explicit){
            if(card.getOpened()[1] && x > 0 && board[x-1][y] instanceof GoalCard){ 
                if(knowGoal[0] == 1 && (x-1) == goalX){
                    return 10;
                } else if(knowGoal[((x-1)/2)-2] != 2){
                    if(branches > 2){
                        return 10;
                    } else{
                        return 9;
                    }
                }
            }
            if(card.getOpened()[3] && x < 12 && board[x+1][y] instanceof GoalCard){ 
                if(knowGoal[0] == 1 && (x+1) == goalX){
                    return 10;
                } else if(knowGoal[((x+1)/2)-2] != 2){
                    if(branches > 2){
                        return 10;
                    } else{
                        return 9;
                    }
                }
            }
            if(card.getOpened()[2] && y > 0 && board[x][y-1] instanceof GoalCard){
                goodY = true;
            } 
            if(card.getOpened()[0] && y < 12 && board[x][y+1] instanceof GoalCard){
                return 10;
            }
            

            if(x > goalX && card.getOpened()[1] && board[x-1][y] == null){ 
                goodX = true;
            }
            if(x < goalX && card.getOpened()[3] && board[x+1][y] == null){
                goodX = true;
            }
            if(y > 10 && card.getOpened()[2] && board[x][y-1] == null){
                goodY = true;
            } 
            if(y < 10 && card.getOpened()[0] && board[x][y+1] == null){
                goodY = true;
            }
            if(Math.abs(goalX - x) > Math.abs(10 - y)){
                if(goodX){
                    cardScore += 5;
                } else if(goodY){
                    cardScore += 4;
                }
                if(goodX && goodY){
                    cardScore += 0.5;
                }
            } else {
                if(goodY){
                    cardScore += 5;
                } else if(goodX){
                    cardScore += 4;
                }
                if(goodX && goodY){
                    cardScore += 0.5;
                }
            }
         
            cardScore += branches-1;
            cardScore += (19-distance)/9;
            double print = (19.0-distance)/9.0;
            //System.out.println("added cardscore for distance: " + print);
            return cardScore;
        }
        if(branches == 0){
            cardScore += 5;
             if(x > goalX && x < 12 && card.getOpened()[3] && board[x+1][y] != null && board[x+1][y].getOpened()[1]){ 
                goodX = true;
            }
            if(x < goalX && x > 0&& card.getOpened()[1] && board[x-1][y] != null && board[x-1][y].getOpened()[3]){
                goodX = true;
            }
            if(y == 11 && card.getOpened()[0] && board[x][y+1] != null && board[x][y+1].getOpened()[2]){
                goodY = true;
            } 
            if(y < 10 && y > 0 && card.getOpened()[2] && board[x][y-1] != null && board[x][y-1].getOpened()[0]){
                goodY = true;
            }
            if(Math.abs(goalX - x) > Math.abs(10 - y)){
                if(goodX){
                    cardScore += 3;
                } else if(goodY){
                    cardScore += 2;
                }
                if(goodX && goodY){
                    cardScore += 0.5;
                }
            } else {
                if(goodY){
                    cardScore += 5;
                } else if(goodX){
                    cardScore += 4;
                }
                if(goodX && goodY){
                    cardScore += 0.5;
                }
            }
            cardScore += (19-distance)/9;
        }
        
        return cardScore;
        
    }
    
    private double calculateDeleteScore(int x, int y, PathCard card, PathCard[][] board, int[] knowGoal) {
        double cardScore = 0;
        int goalX = 6;
        double distance;
        int branches = 0;
        boolean goodX = false;
        boolean goodY = false;
        if(knowGoal[0] == 1){
            for(int i = 1; i < 4; i++){
                if(knowGoal[i] == 1){
                    goalX = (2*i)+2;
                }
            }
        } 
        for(int i = 0; i < 4; i++){
            if(card.getOpened()[i]){
                branches++;
            }
        }
        if(card.getOpened()[1] && x > 0 && board[x-1][y] instanceof GoalCard){ 
            if(knowGoal[0] == 1 && (x-1) == goalX){
                return 10;
            } else if(knowGoal[((x-1)/2)-2] != 2){
                if(branches > 2){
                    return 10;
                } else{
                    return 9;
                }
            }
        }
        if(card.getOpened()[3] && x < 12 && board[x+1][y] instanceof GoalCard){ 
            if(knowGoal[0] == 1 && (x+1) == goalX){
                return 10;
            } else if(knowGoal[((x+1)/2)-2] != 2){
                if(branches > 2){
                    return 10;
                } else{
                    return 9;
                }
            }
        }
        if(card.getOpened()[2] && y > 0 && board[x][y-1] instanceof GoalCard){
            goodY = true;
        } 
        if(card.getOpened()[0] && y < 12 && board[x][y+1] instanceof GoalCard){
            return 10;
        }
        distance = Math.abs(goalX-x) + Math.abs(y-10);
        
        if(x > goalX && card.getOpened()[1] && board[x-1][y] == null){ 
            goodX = true;
        }
        if(x < goalX && card.getOpened()[3] && board[x+1][y] == null){
            goodX = true;
        }
        if(y > 10 && card.getOpened()[2] && board[x][y-1] == null){
            goodY = true;
        } 
        if(y < 10 && card.getOpened()[0] && board[x][y+1] == null){
            goodY = true;
        }
        if(Math.abs(goalX - x) > Math.abs(10 - y)){
            if(goodX){
                cardScore += 5;
            } else if(goodY){
                cardScore += 4;
            }
            if(goodX && goodY){
                cardScore += 0.5;
            }
        } else {
            if(goodY){
                cardScore += 5;
            } else if(goodX){
                cardScore += 4;
            }
            if(goodX && goodY){
                cardScore += 0.5;
            }
        }
        
        cardScore += branches-1;
        cardScore += (19-distance)/9;
        double print = (19.0-distance)/9.0;
        //System.out.println("added cardscore for distance: " + print);
        
        return cardScore;
    }

    private boolean linksToStart(PathCard[][] board, int xCoordinate, int yCoordinate, boolean[][] shadowBoard) {
        boolean[] opened = board[xCoordinate][yCoordinate].getOpened();
        boolean[] sides = board[xCoordinate][yCoordinate].getSides();
        boolean[] links = new boolean[4];
        shadowBoard[xCoordinate][yCoordinate] = true;//remember that recursion already evaluated this tile
        if(xCoordinate == 6 && yCoordinate == 2) {
            return true;
        } else {           
            if(yCoordinate < 12){
                if(sides[0] && board[xCoordinate][yCoordinate+1]!=null && board[xCoordinate][yCoordinate+1].getOpened()[2] && shadowBoard[xCoordinate][yCoordinate+1]==false) {//check if there is a card there              
                    links[0] = linksToStart(board, xCoordinate, yCoordinate+1, shadowBoard);//Card at North
                }
            }
            if(yCoordinate > 0){
                if(sides[2] && board[xCoordinate][yCoordinate-1]!=null && board[xCoordinate][yCoordinate-1].getOpened()[0] && shadowBoard[xCoordinate][yCoordinate-1]==false) {//check if there is a card there           
                    links[1] = linksToStart(board, xCoordinate, yCoordinate-1, shadowBoard);//Card at South
                }  
            }
            if(xCoordinate > 0){
                if(sides[1] && board[xCoordinate-1][yCoordinate]!=null && board[xCoordinate-1][yCoordinate].getOpened()[3] && shadowBoard[xCoordinate-1][yCoordinate]==false) {//check if there is a card there
                    links[2] = linksToStart(board, xCoordinate-1, yCoordinate, shadowBoard);//Card at Right
                }
            }
            if(xCoordinate < 12){
                if(sides[3] && board[xCoordinate+1][yCoordinate]!=null && board[xCoordinate+1][yCoordinate].getOpened()[1] && shadowBoard[xCoordinate+1][yCoordinate]==false) {//check if there is a card there
                    links[3] = linksToStart(board, xCoordinate+1, yCoordinate, shadowBoard);//Card at Left
                }  
            }
        } 
        if(links[0] || links[1] || links[2] || links[3]){
            return true;
        }
        return false;
    }
    
    public int getGoal(int[] knowGoal){
        for(int i = 1; i < 4; i++){
            if(knowGoal[i] == 2){
                return i;
            }
        }
        return 0;
    }
    
    public int knowsGoal(int knowGoal){
        if(knowGoal == 0){
            return 0;
        }
        return ((knowGoal%2) + 1);
    }
    
    public void setExplicit(boolean explicit){
        this.explicit = explicit;
    }
    
    public boolean getExplicit(){
        return explicit;
    }
    
    public void checkExplicit(PathCard[][] board, KripkeModel model, int id){
        int diffX = 10;
        int diffY = 10;
        for(int x = 1; x < 12; x++){
            for(int y = 7; y < 13; y++){
                if(board[x][y] != null && (board[x][y].getOpened()[0] || board[x][y].getOpened()[1] || board[x][y].getOpened()[2] || board[x][y].getOpened()[3])){
                    diffX = Math.min(Math.abs(x-4),Math.min(Math.abs(x-6), Math.abs(x-8)));
                    diffY = Math.abs(y-10);
                    if(diffX + diffY < 4){
                        explicit = true;
                        return;
                    }
                }
            }
        }
        if(model.getSuspected(id)){
            explicit = true;
        }
    }
}
