package saboteur;
import java.util.*;
import java.lang.Math.*;
/**
 *
 * @author s1692933
 */
public class KripkeModel {
    private State[] states;
    private double[][][] relations = new double[5][10][10];//5 Agents with 10 possible states all connected to eachother
    double treshold = 0.0; // how high must the value of a relation be to be considered possible?
   // int[][] trustValues = new int[5][5]; // does agent i trust agent j?
    
    public KripkeModel(int id, boolean saboteur){
        states = new State[10];
        for(int i=0; i < 10; i++){
            switch(i){ //create every possible state
                case (0):   boolean[] case1 = new boolean[]{true,true,false,false,false};
                            states[i] = new State(case1);
                            break;
                case (1):   boolean[] case2 = new boolean[]{true,false,true,false,false};
                            states[i] = new State(case2);
                            break;
                case (2):   boolean[] case3 = new boolean[]{true,false,false,true,false};
                            states[i] = new State(case3);
                            break;
                case (3):   boolean[] case4 = new boolean[]{true,false,false,false,true};
                            states[i] = new State(case4);
                            break;
                case (4):   boolean[] case5 = new boolean[]{false,true,true,false,false};
                            states[i] = new State(case5);
                            break;
                case (5):   boolean[] case6 = new boolean[]{false,true,false,true,false};
                            states[i] = new State(case6);
                            break;
                case (6):   boolean[] case7 = new boolean[]{false,true,false,false,true};
                            states[i] = new State(case7);
                            break;
                case (7):   boolean[] case8 = new boolean[]{false,false,true,true,false};
                            states[i] = new State(case8);
                            break;
                case (8):   boolean[] case9 = new boolean[]{false,false,true,false,true};
                            states[i] = new State(case9);
                            break;
                case (9):   boolean[] case10 = new boolean[]{false,false,false,true,true};
                            states[i] = new State(case10);
                            break;  
            }
        }
        
        for(int o = 0; o < 5; o++) {
            if(o == id && saboteur){
                for(int p = 0;p < 10; p++) {
                    for(int q = 0; q < 10; q++){
                        if(states[q].getAgentRole(id)){
                            // agent o has a relation from state p to state q
                            // threshold = 2 (below that, a state is not considered to be possible
                            relations[o][p][q] = treshold + 2; 
                        } else{
                            relations[o][p][q] = -100; 
                        }
                    } 
                }
            } else {
                for(int p = 0;p < 10; p++) {
                    for(int q = 0; q < 10; q++){
                        if(!states[q].getAgentRole(o)){
                            // agent o has a relation from state p to state q
                            // threshold = 2 (below that, a state is not considered to be possible
                            relations[o][p][q] = treshold + 2; 
                        } else{
                            relations[o][p][q] = -100; 
                        }
                    } 
                }
            }
        }
    }

    public void updateRelations1(int agentPlayed, int change) {
        for(int i = 0; i < 5; i++){
            if(i != agentPlayed){
                for(int p = 0; p < 10; p++) {
                    for(int q = 0; q < 10; q++){
                        if(!states[q].getAgentRole(agentPlayed)){
                            if(change > 0){
                                relations[i][p][q] += change;
                            } else {
                                relations[i][p][q] += (5*change); 
                            }
                        } else {
                            if(change < 0){
                                relations[i][p][q] -= change;
                            } else {
                                relations[i][p][q] -= (5*change);
                            }
                        }
                    } 
                }
            }
        }
    }
    
    public void updateRelations2(int agent1, int agent2, int change) {
        for(int i = 0; i < 5; i++){
            for(int p = 0; p < 10; p++) {
                for(int q = 0; q < 10; q++){
                    //in states where agent1 and agent2 have the same role
                    if(states[q].haveSameRole(agent1,agent2)) { 
                        relations[i][p][q] += change;
                    } else {
                        relations[i][p][q] -= (1.5*change);
                    }
                } 
            }
        }
    }
    
    // update beliefs/relations for all agents in the model of the agent who called this function, as a result of the last turn
    public void update(Turn turn, int agentPlayed, PathCard[][] board, Agent[] agents) {
        // playCard is the action done during the last round (by agentPlayed)   
        Card playCard = turn.getCard();
        int[] target = turn.getTarget(); // target of the playCard (coordinates or another player)
        int agentReceived = target[0];

        if(playCard instanceof DemolishCard) {
            // was the card open (i.e. is there a path)?
            boolean opened = false;
            for(int i = 0; i < 4; i++){
                if(board[target[0]][target[1]].getOpened()[i]){
                    opened = true;
                    i = 4;
                }
            }
            if(opened == false) { //if it was helpful to demolish the card, +1
                updateRelations1(agentPlayed, 1);
                
            } else { // if it was nog helpful: deduction depends on how bad it was
                updateRelations1(agentPlayed, (-1 - closerToGoal(board[target[0]][target[1]], target, board, agents[agentPlayed])));
            }
            
        } else if(playCard instanceof PathCard) {
            PathCard pathCard = (PathCard) playCard;
            updateRelations1(agentPlayed, closerToGoal(pathCard, target, board, agents[agentPlayed]));
            
          

        } else if(playCard instanceof BlockCard) { // Agent agentPlayed blocked agentReceived -> less likely they have the same role
            updateRelations2(agentPlayed,agentReceived,-2);

            
        } else if(playCard instanceof DeblockCard) {
            int agentsBlocked = 0;
            for(int i = 0; i < 5; i++){
                if(agents[i].getBlocked()){
                    agentsBlocked++;
                }
            }
            
            if(agentsBlocked == 1) { //if there was only agent blocked
                // agentPlayed deblocked agentReceived, more likely they have the same role
                updateRelations2(agentPlayed,agentReceived,1);
            } else {
                // agentPlayed deblocked agentReceived, but not the other blocked agents, even more likely they have the same role
                updateRelations2(agentPlayed,agentReceived,2);
                for(int i = 0; i < 5; i++) {// For every other blocked agent:
                    if(i != agentPlayed && i != agentReceived && agents[i].getBlocked()) {
                        updateRelations2(agentPlayed,i,-1); 
                    }
                } 
            }
            
        } else if(playCard instanceof ViewCard) {
            int plus = 0;
            int min = 0;
            boolean know = false;
            boolean someoneKnows = false;
            for(int i = 0; i < 5; i++){
                if(i != agentPlayed && agents[i].knowsGoal((target[0]/2)-2) != 0){
                    someoneKnows = true;
                }
            }
            if(someoneKnows) { //if the goal card Agent agentPlayed looked at, has been looked at before
                for(int i = 0; i < 5; i++){
                    //for all agents that know the goal (besides agentPlayed)
                    if(i != agentPlayed && agents[i].knowsGoal((target[0]/2)-2)!=0){
                        for(int p = 0; p < 10; p++) {
                            for(int q = 0; q < 10; q++){
                                //in states where agent i and agent agentPlayed have the same role
                                if(states[q].haveSameRole(i,agentPlayed)) { 
                                    // did they tell the same thing (more likely that they have the same role) or not?
                                    if(agents[i].knowsGoal((target[0]/2)-2) == agents[agentPlayed].knowsGoal((target[0]/2)-2)){
                                        relations[i][p][q] += 1;
                                    } else {
                                        relations[i][p][q] += -2;
                                    }
                                } else { //in states where agent i and agent agentPlayed have different roles
                                    // did they tell the same thing (less likely that they have the same role) or not?
                                    if(agents[i].knowsGoal((target[0]/2)-2) == agents[agentPlayed].knowsGoal((target[0]/2)-2)){
                                        relations[i][p][q] += -1;
                                    } else {
                                        relations[i][p][q] += 1;
                                    }
                                }
                            } 
                        }
                    } else if(i == agentPlayed){ // agent agentPlayed can also update his beliefs
                        for(int p = 0; p < 10; p++) {
                            for(int q = 0; q < 10; q++){
                                for(int j = 0; j < 5; j++){
                                // did the other agents tell the same as he did? 
                                // then it is more likely that they have the same role, otherwise it is not
                                    if(states[q].haveSameRole(i,j)) { 
                                        if(agents[i].knowsGoal((target[0]/2)-2) == agents[j].knowsGoal((target[0]/2)-2)){
                                            relations[i][p][q] += 1;
                                        } else {
                                            relations[i][p][q] += -2;
                                        }
                                    } else {
                                        if(agents[i].knowsGoal((target[0]/2)-2) == agents[j].knowsGoal((target[0]/2)-2)){
                                            relations[i][p][q] += -1;
                                        } else {
                                            relations[i][p][q] += 1;
                                        }
                                    }
                                }
                            } 
                        }
                    } else { // relations of other agents about agentPlayed and another agent who has seen the goal card, have to be updated too
                        for(int j = 0; j < 5; j++){
                            if(j != i && j != agentPlayed && agents[j].knowsGoal((target[0]/2)-2) != 0){
                                if(agents[j].knowsGoal((target[0]/2)-2) == agents[agentPlayed].knowsGoal((target[0]/2)-2)){
                                    for(int p = 0; p < 10; p++){
                                        for(int q = 0; q < 10; q++){
                                            if(states[q].haveSameRole(j, agentPlayed)) { 
                                                relations[i][p][q] += 1;
                                            } else {
                                                relations[i][p][q] -= 1;
                                            }
                                        }
                                    }
                                } else {
                                    for(int p = 0; p < 10; p++){
                                        for(int q = 0; q < 10; q++){
                                            if(states[q].haveSameRole(j, agentPlayed)) { 
                                                relations[i][p][q] -= 2;
                                            } else {
                                                relations[i][p][q] += 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }     
            }
        }
    }
    
    // miners use second order beliefs to update their knowledge
    // they tend to take over the beliefs of the agent that is least suspicious according to their model
    public void updateSecondOrderBeliefs(int agent) {
        boolean reachable = false;
        double[] miner = getLeastSuspicious(agent);
        if(miner[0] > 1 && (miner[2]-miner[0]) < 2) { // the least suspicious agent needs to be a miner in more than 1 state and a saboteur in less than two states
            for(int i = 0; i < 10; i++){ //iterate over every state
                for(int j = 0; j < 10; j++){ //iterate over every state
                    if(relations[(int)miner[1]][j][i] >= treshold && i!=j){ //determine whether there is a relation between the two states currently iterated over
                        reachable = true;
                        j = 10; // end nested loop, we now know the state is still reachable
                    }
                }
                for(int j = 0; j < 10; j++){ 
                    if(reachable){ 
                        relations[agent][j][i] += 0.5;
                    } else {
                        relations[agent][j][i] -= 0.5;
                    }
                }

                reachable = false;//for next iteration reset reachablity
            }
        }
    }
    
    //Calculate which agent is most suspicious, according to agent 'agent'
    public double[] getSuspicion(int agent){
        return getSuspicion(agent, -1);
    }
    
    //overloading so you can add one other agent to disregard when checking
    public double[] getSuspicion(int agent, int notit){
        int suspicion[] = new int[]{-1,-1,-1};//placeholder for most suspicious agent, level of suspicion and total of reachable worlds will be returned later
        boolean reachable = false; //to determine whether a certain state is accessible
        int counter[] = new int[]{0,0,0,0,0,0}; //counters for every agent + counter for total amount of reachable states
        
        for(int i = 0; i < 10; i++){ //iterate over every state
            for(int j = 0; j < 10; j++){ //iterate over every state
                if(relations[agent][j][i] >= treshold && i!=j){ //determine whether there is a relation between the two states currently iterated over
                    reachable = true; //state is reachable
                    counter[5]++;//up the counter for total amount of reachable states
                    j = 10; // end nested loop, we now know the state is still reachable
                }
            }
            if(reachable){ //only check roles if a state is reachable
                for(int k = 0; k < 5; k++){ //iterate over every agent
                    if(agent!=k && notit!=k && states[i].getAgentRole(k)){//checks if agent k is a saboteur in state i,disregard yourself and extra agent you are not interested in
                        counter[k]++;//raise suspicion for that agent by one
                    }
                }
            }
            reachable = false;//for next iteration reset reachablity
        }
        int j;
        for(int i = 0; i < 5; i++){ //iterate over all counters
            j = (i+agent)%5;// in case there are more players with the same level of suspicion, take the one that comes first after you
            if(counter[j]>suspicion[0]){ //check if agent with id i is more suspicious than currently most suspicious agent
                suspicion[0] = counter[j];//set return variable to highest level of suspicion
                suspicion[1] = j;//set return variable to most suspicious agent
            }
        }
        suspicion[2] = counter[5];
        double returnvalue[] = new double[3];
        for(int i = 0; i < 3; i++){
            returnvalue[i] = (double)suspicion[i];
        }
        return returnvalue;
    }
    
    public double[] getLeastSuspicious(int agent){
        //Calculate which agent is least suspicious, according to agent 'agent'
        return getSuspicion(agent, -1);
    }
    
    //overloading so you can add one other agent to disregard when checking
    public double[] getLeastSuspicious(int agent, int notit){
        int suspicion[] = new int[]{-1,-1,-1};//placeholder for level of 'unsuspicion least suspicious agent, and total of reachable worlds will be returned later
        boolean reachable = false; //to determine whether a certain state is accessible
        int counter[] = new int[]{0,0,0,0,0,0}; //counters for every agent + counter for total amount of reachable states
        
        for(int i = 0; i < 10; i++){ //iterate over every state
            for(int j = 0; j < 10; j++){ //iterate over every state
                if(relations[agent][j][i] >= treshold && i!=j){ //determine whether there is a relation between the two states currently iterated over
                    reachable = true; //state is reachable
                    counter[5]++;//up the counter for total amount of reachable states
                    j = 10; // end nested loop, we now know the state is still reachable
                }
            }
            if(reachable){ //only check roles if a state is reachable
                for(int k = 0; k < 5; k++){ //iterate over every agent
                    if(agent!=k && notit!=k && !states[i].getAgentRole(k)){//checks if agent k is a miner in state i,disregard yourself and extra agent you are not interested in
                        counter[k]++;//raise 'unsuspicion' for that agent by one
                    }
                }
            }
            reachable = false;//for next iteration reset reachablity
        }
        int j;
        for(int i = 0; i < 5; i++){ //iterate over all counters
            j = (i+agent)%5;// in case there are more players with the same level of 'unsuspicion', take the one that comes first after you
            if(counter[j]>suspicion[0]){ //check if agent with id i is more 'unsuspicious' than currently most 'unsuspicious' agent
                suspicion[0] = counter[j];//set return variable to highest level of 'unsuspicion'
                suspicion[1] = j;//set return variable to most 'unsuspicious' agent
            }
        }
        suspicion[2] = counter[5];
        double returnvalue[] = new double[3];
        for(int i = 0; i < 3; i++){
            returnvalue[i] = (double)suspicion[i];
        }
        return returnvalue;
    }
    
    private int closerToGoal(PathCard pathCard,int[] target,PathCard[][] board, Agent agent){
        boolean directionX = false;
        boolean directionY = false;
        int returnValue = 0;
        if((target[1] < 10 && pathCard.getOpened()[0]) || (target[1] > 10 && pathCard.getOpened()[2])){
            directionY = true;
        } else if((target[0] > 8 && pathCard.getOpened()[1]) || (target[0] < 4 && pathCard.getOpened()[3]) || (target[0] > 4 && target[0] < 8)){
            directionX = true;
        } 
        // the path does not bring you closer to the goal cards
        if(!directionX && !directionY){
            return -2;
        }
        // the path card does bring you closer to the goal cards in the y-direction
        if(directionY){
            returnValue += 1;
        }
        
        if(agent.getGoal() == 1){
            // if the opening is in the direction of the specific goal card
            if(target[0] > 4 && pathCard.getOpened()[1]){
                return returnValue + 1;
            }
            if(target[0] < 4 && pathCard.getOpened()[3]){
                return returnValue + 1;
            }
            // if not, but it is in the right y-direction
            if(directionY){
                return 1;
            }
            // he knows where the gold is, but he is not going in that direction
            return -1;
        }
        if(agent.getGoal() == 3){
            // if the opening is in the direction of the specific goal card
            if(target[0] > 8 && pathCard.getOpened()[1]){
                return returnValue + 1;
            }
            if(target[0] < 8 && pathCard.getOpened()[3]){
                return returnValue + 1;
            }
            // if not, but it is in the right y-direction
            if(directionY){
                return 1;
            }
            // he knows where the gold is, but he is not going in that direction
            return -1;
        }
        if(agent.getGoal() == 2){
            // if the opening is in the direction of the specific goal card
            if(target[0] > 6 && pathCard.getOpened()[1]){
                return returnValue + 1;
            }
            if(target[0] < 6 && pathCard.getOpened()[3]){
                return returnValue + 1;
            }
            // if not, but it is in the right y-direction
            if(directionY){
                return 1;
            }
            // he knows where the gold is, but he is not going in that direction
            return -1;
        }
        // agent does not know where the goal is, but is going towards the goal cards
        return 1;
    }
    
    public double[][][] getRelations(){
        return relations;
    }
    
    public boolean getSuspected(int id){
        int noOfAgents = 0;
        int noOfWorldsBad = 0;
        int noOfWorldsGood = 0;
        for(int i = 0; i < 5; i++){
            if(i != id){
                for(int j = 0; j < 10; j++){
                    for(int k = 0; k < 10; k++){
                        if(relations[i][j][k] >= treshold && states[k].getAgentRole(id)){
                            noOfWorldsBad++;
                        } else {
                            noOfWorldsGood++;
                        }
                    }
                }
            }
            if(noOfWorldsGood <= 1&& noOfWorldsBad >= noOfWorldsGood){
                noOfAgents++;
            }
            noOfWorldsGood = 0;
            noOfWorldsBad = 0;
        }
        if(noOfAgents >= 2){
            return true;
        }
        return false;
    }
    
    public State[] getStates(){
        return states;
    }
}
